clc;
clear;
close all;

%% read image
img_id = 1;     % change here to select different images
img_name = ['Pattern',sprintf('%02d',img_id),'.bmp'];
img = double(imread(img_name));

%% image gradient
[Fx, Fy] = gradient(double(img));

%% region of interest
% half subset size
M = 10 : 1 : 200;    

% subset center
[h, w] = size(img);
cx = 0.5 * w;
cy = 0.5 * h;

%% average of squared gradient
maxM = max(M);
img_roi_grad_x = Fx((cx-maxM):(cx+maxM),(cx-maxM):(cx+maxM));
img_grad2_average = mean(mean(img_roi_grad_x.^2));

%% Hessian matrix
h2coeff = zeros(length(M),2);   % first column H; second column E{H}
h6coeff = zeros(length(M),2);
h12coeff = zeros(length(M),2);

for m = 1 : length(M) 
    disp(['Processing ', img_name, ' M = ', num2str(M(m))]);

    % subset points
    x = - M(m) : M(m);
    y = - M(m) : M(m);
    [X, Y] = meshgrid(x, y);

    % intensities and gradients
    I = img((cy-M(m)):(cy+M(m)),(cx-M(m)):(cx+M(m)));
    Ix = Fx((cy-M(m)):(cy+M(m)),(cx-M(m)):(cx+M(m)));
    Iy = Fy((cy-M(m)):(cy+M(m)),(cx-M(m)):(cx+M(m))); 

    % 0-order shape function
    hessian2 = HessianMatrix2(Ix, Iy);
    hessian2_inv = inv(hessian2);
    h2coeff(m,1) = hessian2_inv(1,1) * hessian2(1,1);
    h2coeff(m,2) = ((2*M(m)+1)^2 * img_grad2_average)^-1 * hessian2(1,1);

    % 1-order shape function
    hessian6 = HessianMatrix6(X, Y, Ix, Iy);
    hessian6_inv = inv(hessian6);
    h6coeff(m,1) = hessian6_inv(1,1) * hessian6(1,1);   
    h6coeff(m,2) = ((2*M(m)+1)^2 * img_grad2_average)^-1 * hessian6(1,1);

    % 2-order shape function
    hessian12 = HessianMatrix12(X, Y, Ix, Iy);
    hessian12_inv = inv(hessian12);
    h12coeff(m,1) = hessian12_inv(1,1) * hessian12(1,1); 
%     h12coeff(m,2) = ((2*M(m)+1)^2 * img_grad2_average * ...
%         (4*M(m)*M(m)+4*M(m)-3) / (14*M(m)*M(m)+14*M(m)-3))^-1 * hessian12(1,1);
    h12coeff(m,2) = ((2*M(m)+1)^2 * img_grad2_average * 2 / 7)^-1 * hessian12(1,1);
end 

%% compare Hessian and its ensemble average
figure, hold on;
plot(M,h6coeff(:,1),'s',...
    'MarkerSize',9,...
    'MarkerEdgeColor','none',...
    'MarkerFaceColor',[191,191,255]/255);
plot(M,h6coeff(:,2),'-',...
    'Color',[0,0,255]/255,...
    'LineWidth',2);
line([0 200],[1 1],...
    'LineStyle','--');
plot(M,h12coeff(:,1),'o',...
    'MarkerSize',9,...
    'MarkerEdgeColor','none',...
    'MarkerFaceColor',[255,191,191]/255);
plot(M,h12coeff(:,2),'-',...
    'Color',[255,0,0]/255,...
    'LineWidth',2);
line([0 200],[3.5 3.5],...
    'LineStyle','--');
xlabel('Half Subset Size');
ylabel('Coefficients');
legend('1st Hessian','1st Hessian Ensemble Average','1st Limit=1',...
    '2nd Hessian','2nd Hessian Ensemble Average','2st Limit=3.5');


data = [M', h6coeff, h12coeff];